# -*- coding: utf-8 -*-
from bda.plone.orders.tests import Orders_INTEGRATION_TESTING
from bda.plone.orders.tests import set_browserlayer
from decimal import Decimal
from plone.restapi.interfaces import ISerializeToJson
from zope.component import getMultiAdapter

import unittest


class TestSerializers(unittest.TestCase):
    layer = Orders_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)
        self.count = 1

    def _serializer(self, obj):
        serializer = getMultiAdapter((obj, self.request), ISerializeToJson)
        return serializer

    def _create_raw_record(self):
        from node.ext.zodb import OOBTNode

        record = OOBTNode()
        record.attrs["uid"] = "Unique-Id-{0:04d}".format(self.count)
        self.count += 1
        record.attrs["net"] = Decimal("123.45")
        record.attrs["yesno"] = False
        return record

    def _create_booking(self, order_uid=None):
        record = self._create_raw_record()
        from bda.plone.orders.common import get_bookings_soup

        bookings_soup = get_bookings_soup(self.portal)
        if order_uid:
            record.attrs["order_uid"] = order_uid
        bookings_soup.add(record)
        from bda.plone.orders.common import BookingData

        booking = BookingData(self.portal, booking=record)
        return booking

    def _create_order(self, bookings=1):
        record = self._create_raw_record()
        record.attrs["order_uids"] = []
        for count in range(0, bookings):
            booking = self._create_booking(order_uid=record.attrs["uid"])
            record.attrs["order_uids"].append(booking.uid)
        from bda.plone.orders.common import get_orders_soup

        orders_soup = get_orders_soup(self.portal)
        orders_soup.add(record)
        from bda.plone.orders.common import OrderData

        order = OrderData(self.portal, order=record)
        return order

    def test_booking_serialization(self):
        booking = self._create_booking()
        self.assertDictEqual(
            {"net": 123.45, "uid": u"Unique-Id-0001", "yesno": False},
            self._serializer(booking)(),
        )

    def test_order_serialization(self):
        order = self._create_order()
        self.assertDictEqual(
            {
                "bookings": [
                    {
                        "net": 123.45,
                        "order_uid": "Unique-Id-0001",
                        "uid": "Unique-Id-0002",
                        "yesno": False,
                    }
                ],
                "order": {
                    "net": 123.45,
                    "order_uids": ["Unique-Id-0002"],
                    "uid": "Unique-Id-0001",
                    "yesno": False,
                },
            },
            self._serializer(order)(),
        )



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.intellidatetime import convert
from bda.intellidatetime import DateTimeConversionError
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders import vocabularies as vocabs
from bda.plone.orders.browser.common import customers_form_vocab
from bda.plone.orders.browser.common import salaried_form_vocab
from bda.plone.orders.browser.common import states_form_vocab
from bda.plone.orders.browser.common import Transition
from bda.plone.orders.browser.common import vendors_form_vocab
from bda.plone.orders.browser.dropdown import BaseDropdown
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_order
from bda.plone.orders.common import get_vendor_by_uid
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.common import get_vendors_for
from bda.plone.orders.datamanagers.booking import BookingData
from bda.plone.orders.interfaces import IBuyable
from bda.plone.orders.transitions import do_transition_for
from bda.plone.orders.transitions import transitions_of_main_state
from bda.plone.orders.transitions import transitions_of_salaried_state
from decimal import Decimal
from odict import odict
from Products.CMFPlone.resources import add_bundle_on_request
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from repoze.catalog.query import Contains
from repoze.catalog.query import Eq
from repoze.catalog.query import Ge
from repoze.catalog.query import InRange
from repoze.catalog.query import Le
from yafowil.base import factory
from zExceptions import InternalError
from zope.i18n import translate

import datetime
import json
import plone.api
import uuid
import yafowil.loader  # noqa


###############################################################################
# bookings related dropdowns and transitions
###############################################################################


class BookingsDropdown(BaseDropdown):
    @property
    def booking_data(self):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            vendor_uids = [vendor_uid]
        else:
            vendor_uids = get_vendor_uids_for()
        return BookingData(self.context, booking=self.record, vendor_uids=vendor_uids)


class BookingStateDropdown(BookingsDropdown):
    name = "state"
    css = "dropdown change_booking_state_dropdown"
    action = "bookingstatetransition"
    vocab = vocabs.state_vocab()
    transitions = vocabs.state_transitions_vocab()

    @property
    def value(self):
        return self.booking_data.state

    @property
    def items(self):
        transitions = transitions_of_main_state(self.value)
        return self.create_items(transitions)


class BookingSalariedDropdown(BookingsDropdown):
    name = "salaried"
    css = "dropdown change_booking_salaried_dropdown"
    action = "bookingsalariedtransition"
    vocab = vocabs.salaried_vocab()
    transitions = vocabs.salaried_transitions_vocab()

    @property
    def value(self):
        return self.booking_data.salaried or ifaces.SALARIED_NO

    @property
    def items(self):
        transitions = transitions_of_salaried_state(self.value)
        return self.create_items(transitions)


class BookingTransition(Transition):
    def do_transition(self, uid, transition, vendor_uids):
        booking_data = BookingData(self.context, uid=uid, vendor_uids=vendor_uids)
        do_transition_for(
            booking_data,
            transition=transition,
            context=self.context,
            request=self.request,
        )
        return booking_data.booking


class BookingStateTransition(BookingTransition):
    dropdown = BookingStateDropdown


class BookingSalariedTransition(BookingTransition):
    dropdown = BookingSalariedDropdown


###############################################################################
# bookings views
###############################################################################


class BookingsView(BrowserView):
    table_view_name = "@@bookingstable"

    def __init__(self, context, request):
        super(BookingsView, self).__init__(context, request)
        add_bundle_on_request(request, "bdajax-jquerytools")
        add_bundle_on_request(request, "bdajax-jquerytools-overlay")
        add_bundle_on_request(request, "datatables")
        add_bundle_on_request(request, "bda-plone-orders")

    def bookings_table(self):
        return self.context.restrictedTraverse(self.table_view_name)()

    def __call__(self):
        # check if authenticated user is vendor
        if not get_vendors_for():
            raise Unauthorized
        return super(BookingsView, self).__call__()


class BookingsTable(BrowserView):
    table_template = ViewPageTemplateFile("templates/table.pt")
    table_id = "bdaplonebookings"
    data_view_name = "@@bookingsdata"

    def rendered_table(self):
        return self.table_template(self)

    def render_filter(self):
        filter_widgets = ""

        # vendor areas of current user
        vendors = vendors_form_vocab()
        vendor_selector = None
        # vendor selection, include if more than one vendor
        if len(vendors) > 2:
            vendor_selector = factory(
                "div:label:select",
                name="vendor",
                value=self.request.form.get("vendor", ""),
                props={
                    "vocabulary": vendors,
                    "label": _("filter_for_vendors", default=u"Filter for vendors"),
                },
            )
            filter_widgets += vendor_selector(request=self.request)

        # customers of current user
        customers = customers_form_vocab()
        customer_selector = None
        # customers selection, include if more than one customer
        if len(customers) > 2:
            customer_selector = factory(
                "div:label:select",
                name="customer",
                value=self.request.form.get("customer", ""),
                props={
                    "vocabulary": customers,
                    "label": _("filter_for_customers", default=u"Filter for customers"),
                },
            )
            filter_widgets += customer_selector(request=self.request)

        states = states_form_vocab()
        state_selector = factory(
            "div:label:select",
            name="state",
            value=self.request.form.get("state", ""),
            props={
                "vocabulary": states,
                "label": _("filter_for_state", default=u"Filter for states"),
            },
        )
        filter_widgets += state_selector(request=self.request)

        salaried = salaried_form_vocab()
        salaried_selector = factory(
            "div:label:select",
            name="salaried",
            value=self.request.form.get("salaried", ""),
            props={
                "vocabulary": salaried,
                "label": _("filter_for_salaried", default=u"Filter for salaried state"),
            },
        )
        filter_widgets += salaried_selector(request=self.request)

        # From date filter.
        from_date = factory(
            "div:label:text",
            name="from_date",
            value=self.request.form.get("from_date", ""),
            props={
                "div.class": "date_from_filter",
                "label": _("filter_from_date", default=u"Filter from date"),
            },
        )
        filter_widgets += from_date(request=self.request)

        # To date filter.
        to_date = factory(
            "div:label:text",
            name="to_date",
            value=self.request.form.get("to_date", ""),
            props={
                "div.class": "date_to_filter",
                "label": _("filter_to_date", default=u"Filter to date"),
            },
        )
        filter_widgets += to_date(request=self.request)

        # group orders
        groups = vocabs.groups_vocab()
        group_selector = factory(
            "div:label:select",
            name="group_by",
            value=self.request.form.get("group_by", "buyable"),
            props={
                "div.class": "group_filter",
                "vocabulary": groups,
                "label": _("group_orders_by", default=u"Group orders by"),
            },
        )
        filter_widgets += group_selector(request=self.request)

        return filter_widgets

    def render_dt(self, colname, record):
        value = record.attrs.get(colname, "")
        if value:
            value = value.strftime(DT_FORMAT)
            return value

    def render_email(self, colname, record):
        email = record.attrs.get(colname, "")
        bookings_quantity = self.render_bookings_quantity(colname, record)
        bookings_total_sum = self.render_bookings_total_sum(colname, record)
        value = (
            '<tr class="group_email">'
            '<td colspan="13">'
            + "<p>"
            + email
            + "</p>"
            + "<span>"
            + translate(
                _("bookings_quantity", default=u"Bookings quantity"), self.request
            )
            + ": "
            + str(bookings_quantity)
            + "</span>"
            "<span>"
            + translate(
                _("bookings_total_sum", default=u"Bookings total sum"), self.request
            )
            + ": "
            + str(bookings_total_sum)
            + "</td></tr>"
        )

        return value

    def render_buyable_uid(self, colname, record):
        # this actually gets the title of a buyable_uid
        title = record.attrs.get("title", "")
        bookings_quantity = self.render_bookings_quantity(colname, record)
        bookings_total_sum = self.render_bookings_total_sum(colname, record)
        value = (
            u'<tr class="group_buyable">'
            u'<td colspan="13">'
            + u"<p>"
            + safe_unicode(title)
            + u"</p>"
            + u"<span>"
            + translate(
                _("bookings_quantity", default=u"Bookings quantity"), self.request
            )
            + u": "
            + safe_unicode(bookings_quantity)
            + u"</span>"
            u"<span>"
            + translate(
                _("bookings_total_sum", default=u"Bookings total sum"), self.request
            )
            + u": "
            + bookings_total_sum
            + u"</td></tr>"
        )

        return value

    def render_count(self, colname, record):
        value = record.attrs.get(colname, "")
        unit = record.attrs.get("quantity_unit", "")
        if value:
            value = Decimal(value)
            value = str(value) + " " + unit
            return value

    def render_price_per_unit(self, colname, record):
        currency = record.attrs.get("currency", "")
        price = self._get_price(record)
        if currency and price:
            value = currency + u" {0:.2f}".format(price)
            return value

    def render_sum(self, colname, record):
        currency = record.attrs.get("currency", "")
        count = record.attrs.get("buyable_count", "")
        price = self._get_price(record)
        if currency and price and count:
            count = Decimal(count)
            sum = price * count
            value = currency + u" {0:.2f}".format(sum)
            return value

    def render_bookings_quantity(self, colname, record):
        value = record._v_bookings_quantity
        if value:
            return str(value)

    def render_bookings_total_sum(self, colname, record):
        currency = record.attrs.get("currency", "")
        value = record._v_bookings_total_sum
        if currency and value:
            value = currency + u" {0:.2f}".format(value)
            return value

    def render_name(self, colname, record):
        firstname = safe_unicode(
            self._get_ordervalue("personal_data.firstname", record)
        )
        lastname = safe_unicode(self._get_ordervalue("personal_data.lastname", record))
        return u"{0}, {1}".format(firstname, lastname)

    def render_address(self, colname, record):
        street = safe_unicode(self._get_ordervalue("billing_address.street", record))
        city = safe_unicode(self._get_ordervalue("billing_address.city", record))
        phone = safe_unicode(self._get_ordervalue("personal_data.phone", record))
        email = safe_unicode(record.attrs.get("email", ""))

        return u"{0} {1}<br/>{2}: {3}<br/>{4}: {5}".format(
            street,
            city,
            _("phone", default=u"Phone"),
            phone,
            _("email", default=u"Email"),
            email,
        )

    @property
    def ajaxurl(self):
        return u"{0}/{1}".format(self.context.absolute_url(), self.data_view_name)

    @property
    def columns(self):
        columns = [
            {
                "id": "email",
                "label": _("email", default=u"Email"),
                "renderer": self.render_email,
                "origin": "b",
            },
            {
                "id": "buyable_uid",
                "label": _("buyable_uid", default=u"Buyable Uid"),
                "renderer": self.render_buyable_uid,
                "origin": "b",
            },
            {
                "id": "ordernumber",
                "label": _("ordernumber", default=u"Ordernumber"),
                "origin": "o",
            },
            {
                "id": "created",
                "label": _("booking_date", default=u"Bookingdate"),
                "renderer": self.render_dt,
                "origin": "b",
            },
            {"id": "title", "label": _("Item", default=u"Item"), "origin": "b"},
            {
                "id": "buyable_comment",
                "label": _("booking_comment", default=u"Comment"),
                "origin": "b",
            },
            {
                "id": "name",
                "label": u"{0}, {1}".format(
                    _("firstname", default=u"First Name"),
                    _("lastname", default=u"Last Name"),
                ),
                "renderer": self.render_name,
                "origin": "o",
            },
            {
                "id": "address",
                "label": _("address", default=u"Address"),
                "renderer": self.render_address,
                "origin": "o",
            },
            {
                "id": "price_per_unit",
                "label": _("price_per_unit", default=u"Price per unit"),
                "renderer": self.render_price_per_unit,
                "origin": "b",
            },
            {
                "id": "buyable_count",
                "label": _("count", default=u"Count"),
                "renderer": self.render_count,
                "origin": "b",
            },
            {
                "id": "sum",
                "label": _("sum", default=u"Sum"),
                "renderer": self.render_sum,
                "origin": "b",
            },
            {
                "id": "booking_quantity",
                "label": _("bookings_quantity", default=u"Bookings quantity"),
                "renderer": self.render_bookings_quantity,
                "origin": "b",
            },
            {
                "id": "bookings_total_sum",
                "label": _("bookings_total_sum", default=u"Bookings total sum"),
                "renderer": self.render_bookings_total_sum,
                "origin": "b",
            },
            {
                "id": "salaried",
                "label": _("salaried", default=u"Salaried"),
                "renderer": self.render_salaried,
                "origin": "b",
            },
            {
                "id": "state",
                "label": _("state", default=u"State"),
                "renderer": self.render_state,
                "origin": "b",
            },
        ]
        return columns

    def jsondata(self):
        soup = get_bookings_soup(self.context)
        aaData = list()
        size, result = self.query(soup)

        columns = self.columns
        colnames = [_["id"] for _ in columns]

        def record2list(record, bookings_quantity=None):
            result = list()
            for colname in colnames:
                coldef = self.column_def(colname)
                renderer = coldef.get("renderer")

                if coldef["origin"] == "o":
                    if renderer:
                        value = renderer(colname, record)
                    else:
                        value = self._get_ordervalue(colname, record)
                else:
                    if renderer:
                        value = renderer(colname, record)
                    else:
                        value = record.attrs.get(colname, "")

                result.append(value)
            return result

        for key in result:
            bookings_quantity = 0
            bookings_total_sum = 0
            for record in result[key]:
                bookings_quantity += record.attrs.get("buyable_count") or 0
                bookings_total_sum += self._get_sum(record)
            for record in result[key]:
                record._v_bookings_quantity = bookings_quantity
                record._v_bookings_total_sum = bookings_total_sum
                aaData.append(record2list(record))

        data = {
            "draw": int(self.request.form["draw"]),
            "recordsTotal": size,
            "recordsFiltered": size,
            "data": aaData,
        }

        self.request.response.setHeader(
            "Content-Type", "application/json; charset=utf-8"
        )
        return json.dumps(data)

    def _get_price(self, record):
        """
        returns net + vat price
        """
        net = record.attrs.get("net", "")
        vat_percent = record.attrs.get("vat", "")
        if net and vat_percent:
            net = Decimal(net)
            vat_percent = Decimal(vat_percent)
            vat = 1 + vat_percent / 100
            return net * vat
        return net and Decimal(net) or Decimal(0)

    def _get_sum(self, record):
        """
        returns net + vat * count
        """
        count = record.attrs.get("buyable_count", "")
        price = self._get_price(record)
        if price and count:
            count = Decimal(count)
            sum = price * count
            return Decimal(sum)
        return Decimal(0)

    def _get_ordervalue(self, colname, record):
        """
        helper method to get the values which are saved on the order and not
        on the booking itself.
        """
        order = get_order(self.context, record.attrs.get("order_uid"))
        value = order.attrs.get(colname, "")
        return value

    def slice(self, fullresult):
        start = int(self.request.form["start"])
        length = int(self.request.form["length"])
        count = 0
        for lr in fullresult:
            if count >= start and count < (start + length):
                yield lr
            if count >= (start + length):
                break
            count += 1

    def column_def(self, colname):
        for column in self.columns:
            if column["id"] == colname:
                return column

    def _datetime_checker(self, from_date, to_date):
        # get portal language for datetime locale formating
        # used for quering
        portal = plone.api.portal.get()
        locale = portal.language
        if len(from_date) > 0:
            try:
                from_date = convert(from_date, locale=locale)
            except DateTimeConversionError:
                return None
        if len(to_date) > 0:
            try:
                to_date = convert(to_date, locale=locale)
            except DateTimeConversionError:
                return None
            if (
                isinstance(to_date, datetime.datetime)
                and to_date.hour == 0
                and to_date.minute == 0
            ):
                to_date = to_date.replace(hour=23, minute=59, second=59)

        if isinstance(from_date, str) and isinstance(to_date, str):
            return None
        if isinstance(from_date, datetime.datetime) and isinstance(to_date, str):
            return Ge("created", from_date)
        if isinstance(from_date, str) and isinstance(to_date, datetime.datetime):
            return Le("created", to_date)
        if isinstance(from_date, datetime.datetime) and isinstance(
            to_date, datetime.datetime
        ):
            return InRange("created", from_date, to_date)

    def _text_checker(self, text):
        # used for quering
        if len(text) < 1:
            return None
        return Contains("text", text + "*")

    def _get_buyables_in_context(self):
        catalog = plone.api.portal.get_tool("portal_catalog")
        path = "/".join(self.context.getPhysicalPath())
        brains = catalog(path=path, object_provides=IBuyable.__identifier__)
        for brain in brains:
            yield brain.UID

    def check_modify_order(self, order):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            vendor_uids = [vendor_uid]
            vendor = get_vendor_by_uid(self.context, vendor_uid)
            user = plone.api.user.get_current()
            if not user.checkPermission(permissions.ModifyOrders, vendor):
                return False
        else:
            vendor_uids = get_vendor_uids_for()
            if not vendor_uids:
                return False
        return True

    def query(self, soup):
        self._get_buyables_in_context()
        req_group_id = self.request.get("group_by", "email")
        # if no req group is set
        if len(req_group_id) == 0:
            req_group_id = "email"

        if req_group_id not in vocabs.groups_vocab():
            raise InternalError("Group not allowed!")
        # not pretty but, otherwise there a problems with vocab + soup attrs
        # that need many places to refactor to make it work
        if req_group_id == "buyable":
            req_group_id = "buyable_uid"

        group_index = soup.catalog[req_group_id]

        # TODO: This was the fastest produceable way for me to
        #       have an empty query to start with and concenate the rest.
        queries = []

        # fetch user vendor uids
        vendor_uids = get_vendor_uids_for()
        # filter by given vendor uid or user vendor uids
        vendor_uid = self.request.form.get("vendor")
        if vendor_uid:
            vendor_uid = uuid.UUID(vendor_uid)
            # raise if given vendor uid not in user vendor uids
            if vendor_uid not in vendor_uids:
                raise Unauthorized
            queries.append(Eq("vendor_uid", vendor_uid))

        # filter by customer if given
        customer = self.request.form.get("customer")
        if customer:
            queries.append(Eq("creator", customer))

        # Filter by state if given
        state = self.request.form.get("state")
        if state:
            queries.append(Eq("state", state))

        # Filter by salaried if given
        salaried = self.request.form.get("salaried")
        if salaried:
            queries.append(Eq("salaried", salaried))

        req_from_date = self.request.get("from_date", "")
        req_to_date = self.request.get("to_date", "")
        date_query = self._datetime_checker(req_from_date, req_to_date)
        if date_query:
            queries.append(date_query)

        req_text = self.request.get("search[value]", "")
        text_query = self._text_checker(req_text)
        if text_query:
            queries.append(text_query)

        if queries:
            # TODO: See the TODO above.
            query = queries[0]
            for q in queries[1:]:
                query = query & q
            dummysize, bookings_set = soup.catalog.query(query)
        else:
            booking_uids = soup.catalog["uid"]
            bookings_set = list(booking_uids._rev_index.keys())

        bookings_set = set(bookings_set)

        buyable_index = soup.catalog["buyable_uid"]
        buyables_set = set()
        # get all buyables for the current context/path
        for buyable_uid in self._get_buyables_in_context():
            try:
                buyables_set.update(buyable_index._fwd_index[buyable_uid])
            except KeyError:
                continue
        # filter bookings_set to only match the current context/path
        bookings_set = bookings_set.intersection(buyables_set)

        unique_group_ids = []
        # for each booking which matches the query append the qroup id once
        for group_id, group_booking_ids in group_index._fwd_index.items():
            if bookings_set.intersection(group_booking_ids):
                unique_group_ids.append(group_id)

        res = odict()
        size = len(unique_group_ids)

        # for each unique group id  get the matching group booking ids
        for group_id in self.slice(unique_group_ids):
            res[group_id] = []
            for group_booking_id in group_index._fwd_index[group_id]:
                if group_booking_id in bookings_set:
                    res[group_id].append(soup.get(group_booking_id))

        return size, res

    def render_salaried(self, colname, record):
        if not self.check_modify_order(record):
            booking_data = BookingData(self.context, booking=record)
            return translate(
                vocabs.salaried_vocab()[booking_data.salaried], context=self.request
            )
        return BookingSalariedDropdown(self.context, self.request, record).render()

    def render_state(self, colname, record):
        if not self.check_modify_order(record):
            booking_data = BookingData(self.context, booking=record)
            return translate(
                vocabs.state_vocab()[booking_data.state], context=self.request
            )
        return BookingStateDropdown(self.context, self.request, record).render()

    def __call__(self):
        # check if authenticated user is vendor
        if not get_vendors_for():
            raise Unauthorized
        # disable diazo theming if ajax call
        if "_" in self.request.form:
            self.request.response.setHeader("X-Theme-Disabled", "True")
        return super(BookingsTable, self).__call__()



# -*- coding: utf-8 -*-
from zope.interface import Attribute
from zope.interface import Interface


class IOrderState(Interface):
    """API to handle states of order and booking"""

    context = Attribute(u"content object context of the state to work with")
    uid = Attribute(u"Unique ID of the state")
    state = Attribute(u"state of the order")
    salaried = Attribute(u"Boolean, true if salaried, else false")

    def update_item_stock(booking, old_state, new_state):
        """Change stock according to transition.
        """

    def increase_stock(booking):
        """Increase stock according to booking information"""

    def decrease_stock(booking):
        """Increase stock according to booking information"""

    def reindex_order(order):
        """newly index given order in storage"""

    def reindex_bookings(bookings):
        """newly index given all given bookings in storage"""


class IOrderData(Interface):
    """API to work with an order"""

    order = Attribute(u"raw order data record")
    bookings = Attribute(u"iterable of raw booking data records")
    currency = Attribute(u"currency valid for the order")
    tid = Attribute(u"")
    net = Attribute(u"net sum of the order")
    vat = Attribute(u"vat sum of the order")
    discount_net = Attribute(u"net part of discount applied")
    discount_vat = Attribute(u"vat part of discount applied")
    shipping = Attribute(u"sum of shipping costs")
    total = Attribute(u"total including all costs")


class IBookingData(Interface):
    """API to work with a booking"""

    uid = Attribute(u"nique ID of the booking")


class IInvoiceSender(Interface):
    """Invoice sender information.
    """

    company = Attribute(u"Sender company name")
    companyadd = Attribute(u"Optional Company additional text line")
    firstname = Attribute(u"Sender first name")
    lastname = Attribute(u"Sender last name")
    street = Attribute(u"Streetg")
    zip = Attribute(u"ZIP")
    city = Attribute(u"City")
    country = Attribute(u"Country")
    phone = Attribute(u"Optional phone number")
    email = Attribute(u"Optional email address")
    web = Attribute(u"Optional web address")
    iban = Attribute(u"Banking connection IBAN")
    bic = Attribute(u"Banking connection BIC")



# -*- coding: utf-8 -*-
from zope.interface import Attribute
from zope.interface import Interface


class ITrading(Interface):
    """Trading related information.
    """

    item_number = Attribute(u"Item Number")
    gtin = Attribute(u"Global Trade Item Number")



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.plone.orders.common import acquire_vendor_or_shop_root
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces.markers import IBuyable
from plone import api
from Products.CMFPlone.interfaces import IPloneSiteRoot
from repoze.catalog import query as rq

import uuid


class OrdersManager(object):
    """Manage orders in a given context
    """

    def __init__(self, context, vendor_uid=None):
        self.context = context
        self.vendor_uid = uuid.UUID(vendor_uid) if vendor_uid else None
        self.shop = acquire_vendor_or_shop_root(context)
        self.orders_soup = get_orders_soup(context)
        self.bookings_soup = get_bookings_soup(context)

    def orders_base_query(self):
        """Creates a base query object for the actual context.

        Idea is to use it and extend it by further filters.
        Takes current context and its vendors into account.
        If a vendor_uid is given, this vendor is filtered only.

        raises:
            Unauthorized the given vendor_uid is not part of the
            current context.

        returns:
            repoze catalog query
        """
        vendor_uids = get_vendor_uids_for()
        if self.vendor_uid is None:
            query = rq.Any("vendor_uids", vendor_uids)
        else:
            if self.vendor_uid not in vendor_uids:
                raise Unauthorized
            query = rq.Any("vendor_uids", [self.vendor_uid])
        if not IPloneSiteRoot.providedBy(self.context):
            buyable_uids = self._get_buyable_uids_in_context()
            query = rq.Any("buyable_uids", buyable_uids)
        return query

    def orders(self, query=None, sort_index=None, reverse=None):
        """iterator with length and records of orders

        query (repoze catalog query):
            query to filter on

        If a query is given, the vendor_uid is ignored.
        if neither query nor vendor_uid are given, all orders valid
        for the actual context and its vendors for the are returned.

        returns:
            iterator with first result length, then all following order records.
        """
        if query is not None:
            query = query & self.orders_base_query()
        else:
            query = self.orders_base_query()
        res = self.orders_soup.lazy(
            query, with_size=True, sort_index=sort_index, reverse=reverse
        )
        length = next(res)
        return length, res

    def orders_data(self, query=None, sort_index=None, reverse=None):
        """iterator with length and orderdata

        query (repoze catalog query):
            query to filter on

        If a query is given, the vendor_uid is ignored.
        if neither query nor vendor_uid are given, all orders valid
        for the actual context and its vendors for the are returned.

        returns:
            iterator with first result length, then all following OrderData
        """
        length, lazy_records = self.orders(query=query, sort_index=None, reverse=None)

        def orderdata_generator():
            for lazy_record in lazy_records:
                yield OrderData(self.context, order=lazy_record())

        return length, orderdata_generator()

    def _get_buyable_uids_in_context(self):
        for brain in api.content.find(
            context=self.context, object_provides=IBuyable.__identifier__
        ):
            yield brain.UID

    def remove(self, filter_query):
        """Remove all orders matching the query.

        Don't drink and use this.
        """
        query = self.orders_base_query()
        length, orderdata_iterator = self.orders_data(query=query)
        if length == 0:
            return False
        for orderdata in orderdata_iterator:
            # remove order and its bookings
            orderdata.remove()
        return True



# -*- coding: utf-8 -*-
from bda.plone.orders import safe_encode
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.interfaces import workflow
from bda.plone.payment.interfaces import IPaymentData
from node.utils import instance_property
from persistent.dict import PersistentDict
from repoze.catalog.query import Eq
from zope.interface import implementer

import six


@implementer(IPaymentData)
class PaymentData(object):
    def __init__(self, context):
        self.context = context

    @instance_property
    def order_data(self):
        return OrderData(self.context, uid=self.order_uid)

    @property
    def amount(self):
        amount = "%0.2f" % self.order_data.total
        amount = amount[: amount.index(".")] + amount[amount.index(".") + 1 :]
        return amount

    @property
    def currency(self):
        return self.order_data.currency

    @property
    def description(self):
        order = self.order_data.order
        attrs = order.attrs
        if six.PY2:
            amount = safe_encode(
                "%s %s" % (self.currency, str(round(self.order_data.total, 2)))
            )
        else:
            amount = "%s %s" % (self.currency, str(round(self.order_data.total, 2)))
        description = ", ".join(
            [
                attrs["created"].strftime(DT_FORMAT),
                attrs["personal_data.firstname"],
                attrs["personal_data.lastname"],
                attrs["billing_address.city"],
                amount,
            ]
        )
        return description

    @property
    def ordernumber(self):
        return self.order_data.order.attrs["ordernumber"]

    def annotations(self, ordernumber):
        self.order_uid = self.uid_for(ordernumber)
        attrs = self.order_data.order.attrs
        if "annotations" not in attrs:
            attrs["annotations"] = PersistentDict()
        return attrs["annotations"]

    def uid_for(self, ordernumber):
        soup = get_orders_soup(self.context)
        for order in soup.query(Eq("ordernumber", ordernumber)):
            return str(order.attrs["uid"])

    def data(self, order_uid):
        self.order_uid = order_uid
        return {
            "amount": self.amount,
            "currency": self.currency,
            "description": self.description,
            "ordernumber": self.ordernumber,
        }


def _payment_handling(event, state):
    data = event.data
    order = OrderData(event.context, uid=event.order_uid)
    order.salaried = state
    # XXX: move concrete payment specific changes to bda.plone.payment.
    # Payment may provide a tid attribute to be used here.
    if event.payment.pid == "six_payment":
        order.tid = data["tid"]
    elif event.payment.pid == "stripe_payment":
        order.tid = data["charge_id"]


def payment_success(event):
    _payment_handling(event, workflow.SALARIED_YES)


def payment_failed(event):
    _payment_handling(event, workflow.SALARIED_FAILED)



# -*- coding: utf-8 -*-
from bda.plone.orders.tests import Orders_INTEGRATION_TESTING
from bda.plone.orders.tests import set_browserlayer

import unittest


class TestContacts(unittest.TestCase):
    layer = Orders_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)

    def test_next_contact_id(self):
        from bda.plone.orders import contacts

        soup = contacts.get_contacts_soup()
        contacts.CID_GENERATION["min"] = 10
        contacts.CID_GENERATION["min"] = 12
        self.assertTrue(10 <= contacts.next_contact_id(soup) <= 12)



# -*- coding: utf-8 -*-
from bda.plone.cart import cookie
from bda.plone.cart.cart import get_data_provider
from bda.plone.cart.cartitem import get_item_data_provider
from bda.plone.cart.cartitem import get_item_state
from bda.plone.cart.cartitem import get_item_stock
from bda.plone.cart.interfaces import IShippingItem
from bda.plone.cart.shipping import Shippings
from bda.plone.cart.utils import get_catalog_brain
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.checkout import CheckoutAdapter
from bda.plone.checkout import CheckoutError
from bda.plone.orders import events
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders.common import acquire_vendor_or_shop_root
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.datamanagers.order import calculate_order_state
from bda.plone.orders.datamanagers.order import create_ordernumber
from bda.plone.payment import Payments
from decimal import Decimal
from node.ext.zodb import OOBTNode
from node.utils import instance_property
from plone.uuid.interfaces import IUUID
from repoze.catalog.query import Eq
from souper.soup import Record
from zope.component import queryAdapter
from zope.event import notify

import datetime
import logging
import plone.api
import uuid


logger = logging.getLogger(__name__)


class OrderCheckoutAdapter(CheckoutAdapter):
    @instance_property
    def order(self):
        return Record()

    @property
    def vessel(self):
        return self.order.attrs

    @property
    def items(self):
        # XXX here the ICartItemDataProvider implementation must be used
        return cookie.extract_items(cookie.read(self.request))

    def ordernumber_exists(self, soup, ordernumber):
        return bool(next(soup.query(Eq("ordernumber", ordernumber), with_size=True)))

    def save(self, providers, widget, data):
        super(OrderCheckoutAdapter, self).save(providers, widget, data)
        order = self.order
        # order UUID
        uid = order.attrs["uid"] = uuid.uuid4()
        # manage empty email
        if not order.attrs["personal_data.email"]:
            order.attrs["personal_data.email"] = ""
        # order creator
        creator = None
        member = plone.api.user.get_current()
        if member:
            creator = member.getId()
        order.attrs["creator"] = creator
        # order creation date
        created = datetime.datetime.now()
        order.attrs["created"] = created
        cart_data = get_data_provider(self.context, self.request)
        # payment related information
        if cart_data.total > Decimal(0):
            payment_param = "checkout.payment_selection.payment"
            pid = data.fetch(payment_param).extracted
            payment = Payments(self.context).get(pid)
            order.attrs["payment_method"] = pid
            if payment:
                order.attrs["payment_label"] = payment.label
            else:
                order.attrs["payment_label"] = _("unknown", default=u"Unknown")
        # no payment
        else:
            order.attrs["payment_method"] = "no_payment"
            order.attrs["payment_label"] = _("no_payment", default=u"No Payment")
        # shipping related information
        if cart_data.include_shipping_costs:
            shipping_param = "checkout.shipping_selection.shipping"
            sid = data.fetch(shipping_param).extracted
            shipping = Shippings(self.context).get(sid)
            order.attrs["shipping_method"] = sid
            order.attrs["shipping_label"] = shipping.label
            order.attrs["shipping_description"] = shipping.description
            try:
                order.attrs["shipping_net"] = shipping.net(self.items)
                order.attrs["shipping_vat"] = shipping.vat(self.items)
                order.attrs["shipping"] = (
                    order.attrs["shipping_net"] + order.attrs["shipping_vat"]
                )
            # B/C for bda.plone.shipping < 0.4
            except NotImplementedError:
                order.attrs["shipping_net"] = shipping.calculate(self.items)
                order.attrs["shipping_vat"] = Decimal(0)
                order.attrs["shipping"] = order.attrs["shipping_net"]
        # no shipping
        else:
            order.attrs["shipping_method"] = "no_shipping"
            order.attrs["shipping_label"] = _("no_shipping", default=u"No Shipping")
            order.attrs["shipping_description"] = ""
            order.attrs["shipping_net"] = Decimal(0)
            order.attrs["shipping_vat"] = Decimal(0)
            order.attrs["shipping"] = Decimal(0)
        # create order bookings
        bookings = self.create_bookings(order)
        # set order state, needed for sorting in orders table
        order.attrs["state"] = calculate_order_state(bookings)
        # set order salaried, needed for sorting in orders table
        order.attrs["salaried"] = ifaces.SALARIED_NO
        # lookup booking uids, buyable uids and vendor uids
        booking_uids = list()
        buyable_uids = list()
        vendor_uids = set()
        for booking in bookings:
            booking_uids.append(booking.attrs["uid"])
            buyable_uids.append(booking.attrs["buyable_uid"])
            vendor_uids.add(booking.attrs["vendor_uid"])
        order.attrs["booking_uids"] = booking_uids
        order.attrs["buyable_uids"] = buyable_uids
        order.attrs["vendor_uids"] = list(vendor_uids)
        # cart discount related information
        # XXX: in order to be able to reliably modify orders, cart discount
        #      rules for this order must be stored instead of the actual
        #      calculated discount.
        cart_discount = cart_data.discount(self.items)
        order.attrs["cart_discount_net"] = cart_discount["net"]
        order.attrs["cart_discount_vat"] = cart_discount["vat"]
        # create ordernumber
        orders_soup = get_orders_soup(self.context)
        ordernumber = create_ordernumber()
        while self.ordernumber_exists(orders_soup, ordernumber):
            ordernumber = create_ordernumber()
        order.attrs["ordernumber"] = ordernumber
        # add order
        orders_soup.add(order)
        # add bookings
        bookings_soup = get_bookings_soup(self.context)
        # list containing items where stock threshold has been reached
        stock_threshold_reached_items = list()
        for booking in bookings:
            bookings_soup.add(booking)
            buyable = get_object_by_uid(self.context, booking.attrs["buyable_uid"])
            item_stock = get_item_stock(buyable)
            # no stock applied
            if item_stock is None:
                stock_warning_threshold = None
            else:
                stock_warning_threshold = item_stock.stock_warning_threshold
            if stock_warning_threshold:
                remaining = booking.attrs["remaining_stock_available"]
                # stock threshold has been reached
                if remaining <= stock_warning_threshold:
                    stock_threshold_reached_items.append(booking.attrs)
        if stock_threshold_reached_items:
            event = events.StockThresholdReached(
                self.context,
                self.request,
                order.attrs["uid"],
                stock_threshold_reached_items,
            )
            notify(event)
        # return uid of added order
        return uid

    def create_bookings(self, order):
        ret = list()
        cart_data = get_data_provider(self.context)
        for uid, count, comment in self.items:
            booking = self.create_booking(order, cart_data, uid, count, comment)
            if booking:
                ret.append(booking)
        return ret

    def create_booking(self, order, cart_data, uid, count, comment):
        brain = get_catalog_brain(self.context, uid)
        # brain could be None if uid for item in cookie which no longer exists.
        if not brain:
            msg = u"Item was removed from shop, uid={0}".format(uid)
            logger.warning(msg)
            raise CheckoutError(msg)
        buyable = brain.getObject()
        item_state = get_item_state(buyable, self.request)
        if not item_state.validate_count(count):
            msg = u"Item no longer available {0}".format(buyable.id)
            logger.warning(msg)
            raise CheckoutError(msg)
        item_stock = get_item_stock(buyable)
        # stock not applied, state new
        if item_stock is None:
            available = None
            state = ifaces.STATE_NEW
        # calculate state from stock
        else:
            if item_stock.available is not None:
                item_stock.available -= float(count)
            available = item_stock.available
            state = (
                ifaces.STATE_NEW
                if available is None or available >= 0.0
                else ifaces.STATE_RESERVED
            )
        item_data = get_item_data_provider(buyable)
        vendor = acquire_vendor_or_shop_root(buyable)
        booking = OOBTNode()
        booking.attrs["email"] = order.attrs["personal_data.email"]
        booking.attrs["uid"] = uuid.uuid4()
        booking.attrs["buyable_uid"] = uid
        booking.attrs["buyable_count"] = count
        booking.attrs["buyable_comment"] = comment
        booking.attrs["order_uid"] = order.attrs["uid"]
        booking.attrs["vendor_uid"] = uuid.UUID(IUUID(vendor))
        booking.attrs["creator"] = order.attrs["creator"]
        booking.attrs["created"] = order.attrs["created"]
        booking.attrs["exported"] = False
        booking.attrs["title"] = brain and brain.Title or "unknown"
        booking.attrs["net"] = item_data.net
        booking.attrs["vat"] = item_data.vat
        # XXX: in order to be able to reliably modify bookings, item discount
        #      rules for this booking must be stored instead of the actual
        #      calculated discount.
        booking.attrs["discount_net"] = item_data.discount_net(count)
        booking.attrs["currency"] = cart_data.currency
        booking.attrs["quantity_unit"] = item_data.quantity_unit
        booking.attrs["remaining_stock_available"] = available
        booking.attrs["state"] = state
        booking.attrs["salaried"] = ifaces.SALARIED_NO
        booking.attrs["tid"] = "none"
        shipping_info = queryAdapter(buyable, IShippingItem)
        booking.attrs["shippable"] = shipping_info and shipping_info.shippable
        trading_info = queryAdapter(buyable, ifaces.ITrading)
        booking.attrs["item_number"] = (
            trading_info.item_number if trading_info else None
        )
        booking.attrs["gtin"] = trading_info.gtin if trading_info else None
        return booking



# -*- coding: utf-8 -*-
from plone import api


def reindex_customer_role(context, event):
    """Reindex ``customer_role`` index after local roles changed.
    """
    catalog = api.portal.get_tool(name="portal_catalog")
    catalog.reindexObject(context, idxs=["customer_role"], update_metadata=0)



# -*- coding: utf-8 -*-
from bda.plone.orders.common import get_order
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.text import CatalogTextIndex
from repoze.catalog.query import Eq
from souper.interfaces import ICatalogFactory
from souper.soup import get_soup
from souper.soup import NodeAttributeIndexer
from souper.soup import NodeTextIndexer
from souper.soup import Record
from zope.interface import implementer

import random
import uuid


CID_GENERATION = {
    # valid range for customer ids
    "min": 1,
    "max": 1000000,
    # maximum attempts for creating a new contact id before failure
    "attempts": 10,
}
LOOKUP_QUERY_MAPPING = [
    ("firstname", "personal_data.firstname"),
    ("lastname", "personal_data.lastname"),
    ("zip", "billing_address.zip"),
    ("street", "billing_address.street"),
]


def get_contacts_soup(context):
    return get_soup("bda_plone_orders_contacts", context)


class ContactAttributeIndexer(NodeAttributeIndexer):
    """souper indexer for export contact attributes.
    """

    def __call__(self, context, default):
        if self.attr in context.attrs:
            return context.attrs[self.attr].lower()
        return default


@implementer(ICatalogFactory)
class ContactsCatalogFactory(object):
    def __call__(self, context=None):
        catalog = Catalog()
        uid_indexer = NodeAttributeIndexer("uid")
        catalog[u"uid"] = CatalogFieldIndex(uid_indexer)
        email_indexer = NodeAttributeIndexer("personal_data.email")
        catalog[u"email"] = CatalogFieldIndex(email_indexer)
        cid_indexer = NodeAttributeIndexer("cid")
        catalog[u"cid"] = CatalogFieldIndex(cid_indexer)
        firstname_indexer = ContactAttributeIndexer("personal_data.firstname")
        catalog[u"firstname"] = CatalogFieldIndex(firstname_indexer)
        lastname_indexer = ContactAttributeIndexer("personal_data.lastname")
        catalog[u"lastname"] = CatalogFieldIndex(lastname_indexer)
        zip_indexer = ContactAttributeIndexer("billing_address.zip")
        catalog[u"zip"] = CatalogFieldIndex(zip_indexer)
        street_indexer = ContactAttributeIndexer("billing_address.street")
        catalog[u"street"] = CatalogFieldIndex(street_indexer)
        search_attributes = [
            "personal_data.email",
            "personal_data.firstname",
            "personal_data.lastname",
        ]
        text_indexer = NodeTextIndexer(search_attributes)
        catalog[u"text"] = CatalogTextIndex(text_indexer)
        return catalog


# attributes for extracting contacts
CONTACT_ATTRIBUTES = [
    "personal_data.company",
    "personal_data.email",
    "personal_data.gender",
    "personal_data.firstname",
    "personal_data.phone",
    "personal_data.lastname",
    "billing_address.city",
    "billing_address.country",
    "billing_address.street",
    "billing_address.zip",
    "delivery_address.city",
    "delivery_address.company",
    "delivery_address.country",
    "delivery_address.firstname",
    "delivery_address.street",
    "delivery_address.lastname",
    "delivery_address.zip",
]


def extract_contact(order):
    """Extract export contact from order by ``CONTACT_ATTRIBUTE_MAP`` keys.
    """
    contact = dict()
    for attr in CONTACT_ATTRIBUTES:
        contact[attr] = order.attrs.get(attr, u"").strip()
    return contact


def next_contact_id(soup):
    """Create a unique contact id.
    """
    next_id = None
    for i in range(CID_GENERATION["attempts"]):
        next_id = random.randint(CID_GENERATION["min"], CID_GENERATION["max"])
        res = [r for r in soup.query(Eq("cid", next_id))]
        if not res:
            return next_id
    msg = u"Unable to create unique contact id after {0} attempts.".format(
        CID_GENERATION["attempts"]
    )
    raise ValueError(msg)


def lookup_contact(context, contact):
    """Lookup existing contact from soup by given contact or add to soup
    if inexistent.
    """
    query = None
    for index_key, contact_key in LOOKUP_QUERY_MAPPING:
        query_element = Eq(index_key, contact[contact_key].lower())
        if query is None:
            query = query_element
        else:
            query &= query_element
    soup = get_contacts_soup(context)
    res = soup.query(query)
    record = None
    for rec in res:
        record = rec
        break
    if not record:
        record = Record()
        record.attrs["uid"] = uuid.uuid4()
        record.attrs["cid"] = next_contact_id(soup)
        record.attrs.update(list(contact.items()))
        soup.add(record)
    else:
        record.attrs.update(list(contact.items()))
        soup.reindex([record])
    return record


def save_contact(event):
    """``bda.plone.checkout.interfaces.ICheckoutDone`` subscriber for storing
    order contact to soup.
    """
    order = get_order(event.context, event.uid)
    contact = lookup_contact(event.context, extract_contact(order))
    order.attrs["contact_uid"] = contact.attrs["uid"]



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.plone.ajax import ajax_continue
from bda.plone.ajax import AjaxAction
from bda.plone.cart.utils import ascur
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.checkout import message_factory as _co
from bda.plone.checkout.vocabularies import get_pycountry_name
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders import vocabularies as vocabs
from bda.plone.orders.browser.common import ContentTemplateView
from bda.plone.orders.common import DT_FORMAT
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.common import get_vendor_by_uid
from bda.plone.orders.common import get_vendor_uids_for
from bda.plone.orders.datamanagers.booking import booking_update_comment
from bda.plone.orders.datamanagers.booking import BookingData
from bda.plone.orders.datamanagers.order import OrderData
from bda.plone.orders.transitions import do_transition_for
from plone.memoize import view
from plone.protect.utils import addTokenToUrl
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from repoze.catalog.query import Eq
from yafowil.base import factory
from yafowil.controller import Controller
from zExceptions import BadRequest
from zExceptions import Redirect
from zope.i18n import translate
from zope.security import checkPermission

import plone.api
import uuid


###############################################################################
# order related base classes
# XXX: used by order and invoice views. maybe generalize
###############################################################################


class OrderDataView(BrowserView):
    """Base view for displaying order related data.
    """

    @property
    @view.memoize
    def order_data(self):
        return OrderData(self.context, uid=self.uid)

    @property
    def uid(self):
        # case order uid has been set manually
        try:
            return self._uid
        # fallback to lookup order uid on request
        except AttributeError:
            return self.request.form.get("uid", None)

    @uid.setter
    def uid(self, value):
        self._uid = value

    @property
    def order(self):
        if not self.uid:
            err = _(
                "statusmessage_err_no_order_uid_given",
                default="Cannot show order information because no order uid was given.",  # noqa
            )
            IStatusMessage(self.request).addStatusMessage(err, "error")
            raise Redirect(self.context.absolute_url())
        return dict(self.order_data.order.attrs)

    def country(self, country_id):
        # return value if no id not available i.e. if no dropdown in use
        try:
            return get_pycountry_name(country_id)
        except Exception:
            return country_id


class ProtectedOrderDataView(ContentTemplateView):
    """Protected order data view.

    Expect ordernumber and email to grant access to view details.
    """

    view_template = ViewPageTemplateFile("templates/protected_view.pt")
    content_template = None
    uid = None
    ordernumber = ""
    email = ""
    wrapper_css = "protected_order_data"

    def _form_handler(self, widget, data):
        self.ordernumber = data["ordernumber"].extracted
        self.email = data["email"].extracted

    def render_auth_form(self):
        # Render the authentication form for anonymous users.
        req = self.request
        action = req.getURL()
        ordernumber = self.ordernumber or req.form.get("ordernumber", "")
        email = self.email or req.form.get("email", "")
        form = factory("form", name="content_auth_form", props={"action": action})
        form["ordernumber"] = factory(
            "div:label:error:text",
            value=ordernumber,
            props={
                "label": _("anon_auth_label_ordernumber", default=u"Ordernumber"),
                "div.class": "ordernumber",
                "required": True,
            },
        )
        form["email"] = factory(
            "div:label:error:text",
            value=email,
            props={
                "label": _("anon_auth_label_email", default=u"Email"),
                "div.class": "email",
                "required": True,
            },
        )
        form["submit"] = factory(
            "button",
            props={
                "type": "submit",
                "text": _("anon_auth_label_submit", default=u"Submit"),
                "class": "btn btn-secondary mt-3",
                "handler": self._form_handler,
                "action": "submit",
            },
        )
        controller = Controller(form, req)
        return controller.rendered

    def __call__(self):
        req = self.request
        ordernumber = req.form.get("content_auth_form.ordernumber", None)
        email = req.form.get("content_auth_form.email", None)
        order = None
        errs = []
        if ordernumber and email:
            orders_soup = get_orders_soup(self.context)
            order = orders_soup.query(Eq("ordernumber", ordernumber))
            try:
                # generator should have only one item
                order = next(order)
                try:
                    assert order.attrs["personal_data.email"] == email
                except AssertionError:
                    # Don't raise Unauthorized, as this allows to draw
                    # conclusions on existing ordernumbers
                    order = None
            except StopIteration:
                # order by ordernumber not exists
                order = None
        if not email:
            err = _(
                "anon_auth_err_email",
                default=u"Please provide the email adress you used for "
                u"submitting the order.",
            )
            errs.append(err)
        if not ordernumber:
            err = _(
                "anon_auth_err_ordernumber", default=u"Please provide the ordernumber"
            )
            errs.append(err)
        if email and ordernumber and not order:
            err = _(
                "anon_auth_err_order",
                default=u"No order could be found for the given " u"credentials",
            )
            errs.append(err)
        if not ordernumber and not email:
            # first call of this form
            errs = []
        for err in errs:
            IStatusMessage(self.request).addStatusMessage(err, "error")
        self.uid = order.attrs["uid"] if order else None
        return self.view_template(self)


###############################################################################
# order actions
###############################################################################


class BookingCancel(BrowserView):
    """Cancel booking action.
    """

    def __call__(self):
        booking_uid = self.request.form.get("uid")
        if not booking_uid:
            raise BadRequest("value not given")
        try:
            booking_data = BookingData(self.context, uid=uuid.UUID(booking_uid))  # noqa
            if booking_data.booking is None:
                raise ValueError("invalid value (no booking found)")
            do_transition_for(
                booking_data,
                transition=ifaces.STATE_TRANSITION_CANCEL,
                context=self.context,
                request=self.request,
            )
        except ValueError:
            raise BadRequest("something is wrong with the value")
        order_uid = booking_data.booking.attrs["order_uid"]
        target = u"{}?uid={}".format(self.context.absolute_url(), order_uid)
        action = AjaxAction(
            target=target, name="order", mode="replace", selector=".order_details"
        )
        ajax_continue(self.request, action)


class BookingUpdateComment(BrowserView):
    """Update boking comment action.
    """

    def __call__(self):
        booking_uid = self.request.form.get("uid")
        if not booking_uid:
            raise BadRequest("value not given")
        booking_comment = self.request.form.get("comment")
        try:
            booking_update_comment(self, uuid.UUID(booking_uid), booking_comment)
        except ValueError:
            raise BadRequest("something is wrong with the value")


###############################################################################
# order details
###############################################################################


class OrderViewBase(OrderDataView):
    """Base view for displaying order details.
    """

    @property
    def net(self):
        return ascur(self.order_data.net)

    @property
    def vat(self):
        return ascur(self.order_data.vat)

    @property
    def discount_net(self):
        return ascur(self.order_data.discount_net)

    @property
    def discount_vat(self):
        return ascur(self.order_data.discount_vat)

    @property
    def shipping_title(self):
        # XXX: node.ext.zodb or souper bug with double linked list. figure out
        order = self.order_data.order.attrs
        # order = self.order
        title = translate(order["shipping_label"], context=self.request)
        if order["shipping_description"]:
            title += " (%s)" % translate(
                order["shipping_description"], context=self.request
            )
        return title

    @property
    def shipping_net(self):
        return ascur(self.order_data.shipping_net)

    @property
    def shipping_vat(self):
        return ascur(self.order_data.shipping_vat)

    @property
    def shipping(self):
        # B/C
        return ascur(self.order_data.shipping)

    @property
    def total(self):
        return ascur(self.order_data.total)

    @property
    def currency(self):
        currency = None
        for booking in self.order_data.bookings:
            if currency is None:
                currency = booking.attrs.get("currency")
            if currency != booking.attrs.get("currency"):
                return None
        return currency

    @property
    def listing(self):
        # XXX: discount
        can_cancel_booking = self.can_cancel_booking
        ret = list()
        for booking in self.order_data.bookings:
            obj = get_object_by_uid(self.context, booking.attrs["buyable_uid"])
            state = vocabs.state_vocab()[booking.attrs.get("state")]
            salaried = vocabs.salaried_vocab()[booking.attrs.get("salaried")]
            cancel_target = None
            if can_cancel_booking and state != ifaces.STATE_CANCELLED:
                cancel_target = addTokenToUrl(
                    "{}?uid={}".format(
                        self.context.absolute_url(), booking.attrs["uid"]
                    )
                )
            ret.append(
                {
                    "uid": booking.attrs["uid"],
                    "title": booking.attrs["title"],
                    "url": obj.absolute_url() if obj else None,
                    "cancel_target": cancel_target,
                    "count": booking.attrs["buyable_count"],
                    "net": ascur(booking.attrs.get("net", 0.0)),
                    "discount_net": ascur(float(booking.attrs["discount_net"])),
                    "vat": booking.attrs.get("vat", 0.0),
                    "comment": booking.attrs["buyable_comment"],
                    "quantity_unit": booking.attrs.get("quantity_unit"),
                    "currency": booking.attrs.get("currency"),
                    "state": state,
                    "salaried": salaried,
                }
            )
        return ret

    @property
    def can_modify_order(self):
        return checkPermission("bda.plone.orders.ModifyOrders", self.context)

    @property
    def can_cancel_booking(self):
        return self.can_modify_order and self.order_data.state != ifaces.STATE_CANCELLED

    @property
    def gender(self):
        gender = self.order.get("personal_data.gender", _co("unknown", "Unknown"))
        if gender == "male":
            return _co("male", "Male")
        if gender == "female":
            return _co("female", "Female")
        return gender

    @property
    def payment(self):
        # XXX: node.ext.zodb or souper bug with double linked list. figure out
        order = self.order_data.order.attrs
        # order = self.order
        title = translate(order["payment_label"], context=self.request)
        return title

    @property
    def salaried(self):
        salaried = self.order_data.salaried or ifaces.SALARIED_NO
        return vocabs.salaried_vocab()[salaried]

    @property
    def tid(self):
        tid = [it for it in self.order_data.tid if it != "none"]
        if not tid:
            return _("none", default=u"None")
        return ", ".join(tid)

    @property
    def state(self):
        state = self.order_data.state or ifaces.STATE_NEW
        return vocabs.state_vocab()[state]

    @property
    def created(self):
        value = self.order.get("created", _("unknown", default=u"Unknown"))
        if value:
            value = value.strftime(DT_FORMAT)
        return value

    def exported(self, item):
        return item["exported"] and _("yes", default=u"Yes") or _("no", default=u"No")


class OrderView(OrderViewBase):
    def __call__(self):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            self.vendor_uids = [vendor_uid]
            vendor = get_vendor_by_uid(self.context, vendor_uid)
            user = plone.api.user.get_current()
            if not user.checkPermission(permissions.ModifyOrders, vendor):
                raise Unauthorized
        else:
            self.vendor_uids = get_vendor_uids_for()
            if not self.vendor_uids:
                raise Unauthorized
        return super(OrderView, self).__call__()

    @property
    @view.memoize
    def order_data(self):
        return OrderData(self.context, uid=self.uid, vendor_uids=self.vendor_uids)

    @property
    def ordernumber(self):
        return self.order_data.order.attrs["ordernumber"]


class MyOrderView(OrderViewBase):
    def __call__(self):
        # check if order was created by authenticated user
        user = plone.api.user.get_current()
        if user.getId() != self.order["creator"]:
            raise Unauthorized
        return super(MyOrderView, self).__call__()

    @property
    def ordernumber(self):
        return self.order_data.order.attrs["ordernumber"]


class DirectOrderView(OrderViewBase, ProtectedOrderDataView):
    content_template = ViewPageTemplateFile("templates/order.pt")



# -*- coding: utf-8 -*-
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders.interfaces import IVendor
from plone.app.workflow.interfaces import ISharingPageRole
from zope.interface import implementer


@implementer(ISharingPageRole)
class VendorRole(object):
    title = _(u"title_vendor_role", default=u"Can Sell")
    required_permission = permissions.DelegateVendorRole
    required_interface = IVendor


@implementer(ISharingPageRole)
class CustomerRole(object):
    title = _(u"title_customer_role", default=u"Can Buy")
    required_permission = permissions.DelegateCustomerRole
    required_interface = None



# -*- coding: utf-8 -*-
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

import six.moves.urllib.parse


class BaseDropdown(object):
    """Base class for order related dropdown menus.
    """

    render = ViewPageTemplateFile("templates/dropdown.pt")
    name = ""
    css = "dropdown"
    action = ""
    vocab = {}
    transitions = {}
    value = ""

    def __init__(self, context, request, record):
        self.context = context
        self.request = request
        self.record = record

    def create_items(self, transitions):
        """create and return available transitions for dropdown
        """
        ret = list()
        url = self.context.absolute_url()
        params = {"uid": str(self.record.attrs["uid"])}
        vendor = self.request.form.get("vendor", "")
        if vendor:
            params["vendor"] = vendor
        for transition in transitions:
            params["transition"] = transition
            target = "?".join([url, six.moves.urllib.parse.urlencode(params)])
            ret.append({"title": self.transitions[transition], "target": target})
        return ret

    @property
    def identifyer(self):
        return "%s-%s" % (self.name, str(self.record.attrs["uid"]))

    @property
    def ajax_action(self):
        return "%s:#%s-%s:replace" % (
            self.action,
            self.name,
            str(self.record.attrs["uid"]),
        )

    @property
    def items(self):
        raise NotImplementedError(
            u"Abstract Dropdown does not implement " u"``items``."
        )



# -*- coding: utf-8 -*-
from zope.interface import Attribute
from zope.interface import Interface


class INotificationSettings(Interface):
    """Mail notification settings.
    """

    admin_email = Attribute(u"Shop admin email address")
    admin_name = Attribute(u"Shop admin name")


class IItemNotificationText(Interface):
    """Notification text for items in order.
    """

    order_text = Attribute(u"Text sent after successful checkout for item")
    overbook_text = Attribute(
        u"Text sent after successful checkout for item if stock overbooked"
    )


class IGlobalNotificationText(Interface):
    """Notification text per order.
    """

    global_order_text = Attribute(u"Text sent after successful checkout for order")
    global_overbook_text = Attribute(
        u"Text sent after successful checkout for order if stock overbooked"
    )


class IPaymentText(Interface):
    """Payment related order notification text.
    """

    def payment_text(payment):
        """Text sent after successful checkout for payment
        """



# -*- coding: utf-8 -*-
from bda.plone.orders import interfaces as ifaces
from bda.plone.orders import message_factory as _
from bda.plone.orders.common import get_all_vendors
from bda.plone.orders.common import get_order
from bda.plone.orders.common import get_vendor_order_uids_for
from bda.plone.orders.common import get_vendors_for
from plone.i18n.normalizer.base import baseNormalize
from plone.uuid.interfaces import IUUID
from Products.CMFPlone.utils import safe_unicode
from zope.component.hooks import getSite

import plone.api


def state_vocab():
    vocab = {
        ifaces.STATE_NEW: _("new", default=u"New"),
        ifaces.STATE_PROCESSING: _("processing", default=u"Processing"),
        ifaces.STATE_FINISHED: _("finished", default=u"Finished"),
        ifaces.STATE_CANCELLED: _("cancelled", default=u"Cancelled"),
        ifaces.STATE_RESERVED: _("reserved", default=u"Reserved"),
        ifaces.STATE_MIXED: _("mixed", default=u"Mixed"),
    }
    return vocab


def state_transitions_vocab():
    vocab = {
        ifaces.STATE_TRANSITION_RENEW: _("renew", default=u"Renew"),
        ifaces.STATE_TRANSITION_PROCESS: _("process", default=u"Process"),
        ifaces.STATE_TRANSITION_FINISH: _("finish", default=u"Finish"),
        ifaces.STATE_TRANSITION_CANCEL: _("cancel", default=u"Cancel"),
    }
    return vocab


def salaried_vocab():
    vocab = {
        ifaces.SALARIED_YES: _("yes", default=u"Yes"),
        ifaces.SALARIED_NO: _("no", default=u"No"),
        ifaces.SALARIED_FAILED: _("failed", default=u"Failed"),
        ifaces.SALARIED_MIXED: _("mixed", default=u"Mixed"),
    }
    return vocab


def salaried_transitions_vocab():
    vocab = {
        ifaces.SALARIED_TRANSITION_SALARIED: _(
            "mark_salaried", default=u"Mark salaried"
        ),
        ifaces.SALARIED_TRANSITION_OUTSTANDING: _(
            "mark_outstanding", default=u"Mark outstanding"
        ),
    }
    return vocab


def groups_vocab():
    """
    used for grouping the bookings table
    """
    vocab = {
        "email": _("email", default=u"Email"),
        "buyable": _("buyable", default=u"Buyable"),
    }
    return vocab


def all_vendors_vocab():
    """Vocabulary for all vendor areas by uuid.
    """
    vendors = get_all_vendors()
    vocab = [
        (
            IUUID(vendor),
            u"{0} ({1})".format(
                safe_unicode(vendor.Title()), vendor.absolute_url_path()
            ),
        )
        for vendor in vendors
    ]
    return vocab


def vendors_vocab_for(user=None):
    """Vendors vocabulary for given or currently authenticated user.
    """
    vendors = get_vendors_for(user=user)
    vocab = [
        (
            IUUID(vendor),
            u"{0} ({1})".format(
                safe_unicode(vendor.Title()), vendor.absolute_url_path()
            ),
        )
        for vendor in vendors
    ]
    return vocab


def customers_vocab_for(user=None):
    """Customers vocabulary for given or currently authenticated user.
    """
    # XXX: expect context as argument
    context = getSite()
    order_uids = get_vendor_order_uids_for(context, user=user)
    res = set(get_order(context, uid).attrs["creator"] for uid in order_uids)
    vocab = []
    for creator in res:
        if not creator:
            # Development edge case: creator might be None
            continue
        customer = plone.api.user.get(userid=creator)

        email = None
        name = None
        if customer:
            # soft dep on bda.plone.shop
            first = safe_unicode(customer.getProperty("firstname", ""))
            last = safe_unicode(customer.getProperty("lastname", ""))
            email = safe_unicode(customer.getProperty("email", ""))
            # fallback
            full = safe_unicode(customer.getProperty("fullname", ""))
            name = u"{0}, {1}".format(last, first) if (first or last) else full

        if email and name:
            title = u"{0} ({1}) - {2}".format(name, creator, email)
        else:
            title = creator
        vocab.append((creator, title))

    # Sort the vocab by title, normalized like sortable_title
    vocab = sorted(vocab, key=lambda x: baseNormalize(x[1]).lower())

    return vocab



# -*- coding: utf-8 -*-
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.interfaces.markers import IBuyable
from decimal import Decimal
from plone.uuid.interfaces import IUUID
from repoze.catalog.query import Eq


class BuyableData(object):
    def __init__(self, context):
        if not IBuyable.providedBy(context):
            raise ValueError("Given context is not IBuyable")
        self.context = context

    def item_ordered(self, state=[]):
        """Return total count buyable item was ordered.
        """
        context = self.context
        bookings_soup = get_bookings_soup(context)
        order_bookings = dict()
        for booking in bookings_soup.query(Eq("buyable_uid", IUUID(context))):
            bookings = order_bookings.setdefault(booking.attrs["order_uid"], list())
            bookings.append(booking)
        count = Decimal("0")
        for order_uid, bookings in order_bookings.items():
            for booking in bookings:
                if state and booking.attrs["state"] not in state:
                    continue
                count += booking.attrs["buyable_count"]
        return count



# -*- coding: utf-8 -*-
from bda.plone.orders import events
from bda.plone.orders.datamanagers.booking import BookingData
from bda.plone.orders.interfaces import workflow
from zope.event import notify


"""
State transitions and stock change

origin state    target state    stock change
============================================
NEW             PROCESSING      0
NEW             FINISHED        0
NEW             CANCELLED       +1
RESERVED        PROCESSING      -1
RESERVED        FINISHED        -1
RESERVED        CANCELLED       0
PROCESSING      FINISH          0
PROCESSING      CANCEL          +1
PROCESSING      RENEW           0
FINISHED        RENEW           0
CANCELLED       RENEW           -1

"""


def transitions_of_main_state(state):
    """List of transitions for a given orders or bookings main state
    """
    transitions = list()
    if state in [workflow.STATE_NEW, workflow.STATE_RESERVED]:
        transitions = [
            workflow.STATE_TRANSITION_PROCESS,
            workflow.STATE_TRANSITION_FINISH,
            workflow.STATE_TRANSITION_CANCEL,
        ]
    elif state == workflow.STATE_MIXED:
        transitions = [
            workflow.STATE_TRANSITION_PROCESS,
            workflow.STATE_TRANSITION_FINISH,
            workflow.STATE_TRANSITION_CANCEL,
            workflow.STATE_TRANSITION_RENEW,
        ]
    elif state == workflow.STATE_PROCESSING:
        transitions = [
            workflow.STATE_TRANSITION_FINISH,
            workflow.STATE_TRANSITION_CANCEL,
            workflow.STATE_TRANSITION_RENEW,
        ]
    elif state is not None:
        transitions = [workflow.STATE_TRANSITION_RENEW]
    else:  # empty dropdown
        transitions = []
    return transitions


def transitions_of_salaried_state(state):
    """List of transitions for a given orders or bookings salaried state
    """
    transitions = []
    if state == workflow.SALARIED_YES:
        transitions = [workflow.SALARIED_TRANSITION_OUTSTANDING]
    elif state == workflow.SALARIED_NO:
        transitions = [workflow.SALARIED_TRANSITION_SALARIED]
    else:
        transitions = [
            workflow.SALARIED_TRANSITION_OUTSTANDING,
            workflow.SALARIED_TRANSITION_SALARIED,
        ]
    return transitions


def do_transition_for(order_state, transition, context=None, request=None):
    """Do any transition for given ``OrderState`` implementation.

    This mixes main state and salaried!
    """

    def _set_state(
        data,
        state_value,
        state_attr="state",
        event_class=None,
        event_emit_on_last=False,
    ):
        #
        #                             Case BookingData
        #                    Case OrderData    |
        # BookingData or OrderData    |        |
        #                   V         V        V
        bookings = getattr(data, "bookings", [data])
        bookings = list(bookings)  # make list out of generator to have a len
        bookings_len = len(bookings)
        for cnt, booking_data in enumerate(bookings):

            if not isinstance(booking_data, BookingData):
                # Case OrderData
                # It's a booking record from iterating over ``bookings`` in
                # OrderData. We have to factorize a BookingData object now.
                booking_data = BookingData(context=context, booking=booking_data)

            # Set state. This includes updates via it's setter method (E.g.
            # OrderData state and item_stock).
            setattr(booking_data, state_attr, state_value)

            # Optionally send out event.
            # May include sending out a notification mail.
            if event_class and (
                cnt == bookings_len - 1
                or not event_emit_on_last  # event_emit_on_last  # or emit always
            ):
                booking_attrs = dict(list(booking_data.booking.attrs.items()))
                event = event_class(
                    context=context,
                    request=request,
                    order_uid=booking_attrs["order_uid"],
                    booking_attrs=booking_attrs,
                )
                notify(event)

    if transition == workflow.SALARIED_TRANSITION_SALARIED:
        _set_state(order_state, workflow.SALARIED_YES, "salaried")

    elif transition == workflow.SALARIED_TRANSITION_OUTSTANDING:
        _set_state(order_state, workflow.SALARIED_NO, "salaried")

    elif transition == workflow.STATE_TRANSITION_RENEW:
        _set_state(
            order_state,
            workflow.STATE_NEW,
            event_class=events.OrderSuccessfulEvent,
            event_emit_on_last=True,
        )

    elif transition == workflow.STATE_TRANSITION_PROCESS:
        event_class = None
        if order_state.state == workflow.STATE_RESERVED:
            event_class = events.BookingReservedToOrderedEvent
        _set_state(
            order_state, workflow.STATE_PROCESSING, event_class=event_class
        )  # noqa

    elif transition == workflow.STATE_TRANSITION_FINISH:
        event_class = None
        if order_state.state == workflow.STATE_RESERVED:
            event_class = events.BookingReservedToOrderedEvent
        _set_state(
            order_state, workflow.STATE_FINISHED, event_class=event_class
        )  # noqa

    elif transition == workflow.STATE_TRANSITION_CANCEL:
        _set_state(
            order_state,
            workflow.STATE_CANCELLED,
            event_class=events.BookingCancelledEvent,
        )

    else:
        raise ValueError(u"Invalid transition: %s" % transition)



# -*- coding: utf-8 -*-
from zope.interface import Attribute
from zope.interface import Interface


class IOrderSuccessfulEvent(Interface):
    """Checkout related event.
    """

    context = Attribute(u"Context in which this event was triggered.")
    request = Attribute(u"Current request.")
    order_uid = Attribute(u"UUID of Order")
    booking_attrs = Attribute(u"Dict of attributes of the cancelled booking.")


class IBookingCancelledEvent(Interface):
    """Checkout related event.
    """

    context = Attribute(u"Context in which this event was triggered.")
    request = Attribute(u"Current request.")
    order_uid = Attribute(u"UUID of Order")
    booking_attrs = Attribute(u"Dict of attributes of the cancelled booking.")


class IBookingReservedToOrderedEvent(Interface):
    """Checkout related event.
    """

    context = Attribute(u"Context in which this event was triggered.")
    request = Attribute(u"Current request.")
    order_uid = Attribute(u"UUID of Order")
    booking_attrs = Attribute(u"Dict of attributes of the cancelled booking.")


class IStockThresholdReached(Interface):
    """Checkout related event.
    """

    context = Attribute(u"Context in which this event was triggered.")
    request = Attribute(u"Current request.")
    order_uid = Attribute(u"UUID of Order")
    stock_threshold_reached_items = Attribute(
        u"List of items that are " u"getting out of stock."
    )



# -*- coding: utf-8 -*-
from bda.plone.cart.cartitem import get_item_stock
from bda.plone.cart.utils import get_object_by_uid
from bda.plone.orders.common import get_bookings_soup
from bda.plone.orders.common import get_orders_soup
from bda.plone.orders.common import is_billable_booking
from bda.plone.orders.interfaces import IOrderState
from bda.plone.orders.interfaces import workflow
from node.utils import instance_property
from zope.interface import implementer


@implementer(IOrderState)
class BaseState(object):
    context = None

    @instance_property
    def orders_soup(self):
        return get_orders_soup(self.context)

    @instance_property
    def bookings_soup(self):
        return get_bookings_soup(self.context)

    def reindex_order(self, order):
        self.orders_soup.reindex(records=[order])

    def reindex_bookings(self, bookings):
        self.bookings_soup.reindex(records=bookings)

    @property
    def state(self):
        raise NotImplementedError("Abstract OrderState does not implement state")

    @state.setter
    def state(self, value):
        raise NotImplementedError("Abstract OrderState does not implement state.setter")

    @property
    def salaried(self):
        raise NotImplementedError("Abstract OrderState does not implement salaried")

    @salaried.setter
    def salaried(self, value):
        raise NotImplementedError(
            "Abstract OrderState does not implement salaried.setter"
        )

    def update_item_stock(self, booking, old_state, new_state):
        """Change stock according to transition. See table in transitions.py
        """
        if old_state == new_state:
            return
        # XXX: fix stock item available??
        if old_state == workflow.STATE_NEW:
            if new_state == workflow.STATE_CANCELLED:
                self.increase_stock(booking)
        elif old_state == workflow.STATE_RESERVED:
            if new_state in (workflow.STATE_PROCESSING, workflow.STATE_FINISHED):
                self.decrease_stock(booking)
        elif old_state == workflow.STATE_PROCESSING:
            if new_state == workflow.STATE_CANCELLED:
                self.increase_stock(booking)
        elif old_state == workflow.STATE_FINISHED:
            if new_state == workflow.STATE_NEW:
                # do nothing
                pass
        elif old_state == workflow.STATE_CANCELLED:
            if new_state == workflow.STATE_NEW:
                self.decrease_stock(booking)

    def increase_stock(self, booking):
        obj = get_object_by_uid(self.context, booking.attrs["buyable_uid"])
        # object no longer exists
        if not obj:
            return
        stock = get_item_stock(obj)
        # if stock.available is None, no stock information used
        if stock.available is not None:
            stock.available += float(booking.attrs["buyable_count"])

    def decrease_stock(self, booking):
        obj = get_object_by_uid(self.context, booking.attrs["buyable_uid"])
        # object no longer exists
        if not obj:
            return
        stock = get_item_stock(obj)
        # if stock.available is None, no stock information used
        if stock.available is not None:
            # TODO: ATTENTION: here might get removed more than available..?
            stock.available -= float(booking.attrs["buyable_count"])


def _calculate_order_attr_from_bookings(bookings, attr, mixed_value):
    ret = None
    for booking in bookings:
        val = booking.attrs[attr]
        if ret and ret != val:
            ret = mixed_value
            break
        else:
            ret = val
    return ret


def calculate_order_state(bookings):
    return _calculate_order_attr_from_bookings(bookings, "state", workflow.STATE_MIXED)


def calculate_order_salaried(bookings):
    return _calculate_order_attr_from_bookings(
        list(filter(is_billable_booking, bookings)), "salaried", workflow.SALARIED_MIXED
    )



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.plone.cart.utils import ascur
from bda.plone.orders import message_factory as _
from bda.plone.orders.browser.order import OrderDataView
from bda.plone.orders.browser.order import ProtectedOrderDataView
from bda.plone.orders.common import DT_FORMAT_SHORT
from bda.plone.orders.interfaces import IInvoiceSender
from Products.CMFPlone.utils import safe_unicode
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

import plone.api


###############################################################################
# invoice
###############################################################################


class InvoiceViewBase(OrderDataView):
    invoice_prefix = u""

    @property
    def sender(self):
        # XXX: right now we provide a global sender for invoices. in order to
        #      provide vendors there's more to do. there need to be an option
        #      whether the invoice gets splitted to specific vendors and if so
        #      multiple invoices need to be available for one order. this also
        #      needs to be reflected in order emails. if there is one billing
        #      point even if there are multiple vendors for one order the
        #      global sender address is used (which is actually the current
        #      and only behavior).
        sender = IInvoiceSender(self.context)
        return {
            "company": sender.company,
            "companyadd": sender.companyadd,
            "firstname": sender.firstname,
            "lastname": sender.lastname,
            "street": sender.street,
            "zip": sender.zip,
            "city": sender.city,
            "country": sender.country,
            "phone": sender.phone,
            "email": sender.email,
            "web": sender.web,
            "iban": sender.iban,
            "bic": sender.bic,
        }

    @property
    def invoice_number(self):
        return "{}{}".format(self.invoice_prefix, self.order["ordernumber"])

    @property
    def created(self):
        value = self.order.get("created", _("unknown", default=u"Unknown"))
        if value:
            value = value.strftime(DT_FORMAT_SHORT)
        return value

    @property
    def listing(self):
        ret = list()
        for booking in self.order_data.bookings:
            data = dict()
            data["title"] = safe_unicode(booking.attrs["title"])
            data["item_number"] = safe_unicode(booking.attrs["item_number"])
            data["comment"] = safe_unicode(booking.attrs["buyable_comment"])
            data["currency"] = safe_unicode(booking.attrs["currency"])
            data["count"] = booking.attrs["buyable_count"]
            data["net"] = booking.attrs.get("net", 0.0)
            data["vat"] = booking.attrs.get("vat", 0.0)
            data["discount_net"] = float(booking.attrs["discount_net"])
            data["quantity_unit"] = booking.attrs.get("quantity_unit")
            ret.append(data)
        return ret

    def summary(self):
        order_data = self.order_data
        data = dict()
        data["currency"] = order_data.currency
        cart_net = order_data.net
        data["cart_net"] = cart_net
        cart_vat = order_data.vat
        data["cart_vat"] = cart_vat
        discount_net = order_data.discount_net
        data["discount_net"] = discount_net
        discount_vat = order_data.discount_vat
        data["discount_vat"] = discount_vat
        discount_total = discount_net + discount_vat
        data["discount_total"] = discount_total
        shipping_net = order_data.shipping_net
        data["shipping_net"] = shipping_net
        shipping_vat = order_data.shipping_vat
        data["shipping_vat"] = shipping_vat
        shipping_total = shipping_net + shipping_vat
        data["shipping_total"] = shipping_total
        cart_total = order_data.total
        data["cart_total"] = cart_total
        return data

    def ascur(self, val):
        return ascur(val)


class InvoiceView(InvoiceViewBase):
    """Invoice view.
    """


class MyInvoiceView(InvoiceViewBase):
    """My invoice view.
    """

    def __call__(self):
        # check if order was created by authenticated user
        user = plone.api.user.get_current()
        if user.getId() != self.order["creator"]:
            raise Unauthorized
        return super(MyInvoiceView, self).__call__()


class DirectInvoiceView(InvoiceViewBase, ProtectedOrderDataView):
    """Direct Invoice view.
    """

    content_template = ViewPageTemplateFile("templates/invoice.pt")



# -*- coding: utf-8 -*-
from AccessControl import Unauthorized
from bda.plone.orders import message_factory as _
from bda.plone.orders import permissions
from bda.plone.orders import vocabularies as vocabs
from bda.plone.orders.common import get_vendor_by_uid
from bda.plone.orders.common import get_vendor_uids_for
from Products.Five import BrowserView
from zope.i18n import translate
from zope.i18nmessageid import Message

import plone.api


###############################################################################
# utils
###############################################################################


def vendors_form_vocab():
    vendors = vocabs.vendors_vocab_for()
    return [("", _("all", default="All"))] + vendors


def customers_form_vocab():
    customers = vocabs.customers_vocab_for()
    return [("", _("all", default="All"))] + customers


def states_form_vocab():
    states = vocabs.state_vocab()
    return [("", _("all", default="All"))] + list(states.items())


def salaried_form_vocab():
    salaried = vocabs.salaried_vocab()
    return [("", _("all", default="All"))] + list(salaried.items())


class Translate(object):
    def __init__(self, request):
        self.request = request

    def __call__(self, msg):
        if not isinstance(msg, Message):
            return msg
        return translate(msg, context=self.request)


class Transition(BrowserView):
    dropdown = None

    @property
    def vendor_uids(self):
        vendor_uid = self.request.form.get("vendor", "")
        if vendor_uid:
            vendor_uids = [vendor_uid]
            vendor = get_vendor_by_uid(self.context, vendor_uid)
            user = plone.api.user.get_current()
            if not user.checkPermission(permissions.ModifyOrders, vendor):
                raise Unauthorized
        else:
            vendor_uids = get_vendor_uids_for()
            if not vendor_uids:
                raise Unauthorized
        return vendor_uids

    def __call__(self):
        uid = self.request["uid"]
        transition = self.request["transition"]
        vendor_uids = self.vendor_uids
        record = self.do_transition(uid, transition, vendor_uids)
        return self.dropdown(self.context, self.request, record).render()

    def do_transition(self, uid, transition, vendor_uids):
        raise NotImplementedError()


###############################################################################
# content view base classes
###############################################################################


class ContentTemplateView(BrowserView):
    """View mixin for rendering content from dedicated template.
    """

    content_template = None

    def render_content(self):
        return self.content_template(self)



# -*- coding: utf-8 -*-
from bda.plone.orders.common import acquire_vendor_or_shop_root
from bda.plone.orders.interfaces import IVendor
from bda.plone.orders.tests import Orders_INTEGRATION_TESTING
from bda.plone.orders.tests import set_browserlayer
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.interface import alsoProvides

import unittest


class TestOrders(unittest.TestCase):
    layer = Orders_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)


class DummyContext(dict):
    __parent__ = None

    def __nonzero__(self):
        return True

    def __setitem__(self, key, val):
        assert isinstance(val, DummyContext)
        val.__parent__ = self
        super(DummyContext, self).__setitem__(key, val)


class TestOrdersUnit(unittest.TestCase):
    def setUp(self):
        root = DummyContext()
        root["sub1"] = DummyContext()
        root["sub1"]["subsub1"] = DummyContext()
        root["sub2"] = DummyContext()

        alsoProvides(root, IPloneSiteRoot)
        alsoProvides(root["sub1"], IVendor)
        self.root = root

    def test_acquire_vendor_or_shop_root(self):
        root = self.root
        self.assertEqual(
            acquire_vendor_or_shop_root(root["sub1"]["subsub1"]), root["sub1"]
        )
        self.assertEqual(acquire_vendor_or_shop_root(root["sub2"]), root)



# -*- coding: utf-8 -*-
from bda.plone.cart.interfaces import ICartItem
from bda.plone.checkout.interfaces import ICheckoutExtensionLayer
from bda.plone.discount.interfaces import IDiscountSettingsEnabled


class IBuyable(ICartItem, IDiscountSettingsEnabled):
    """Marker for buyable item.

    Item is buyable.
    """


class IOrdersExtensionLayer(ICheckoutExtensionLayer):
    """Browser layer request marker indicating bda.plone.orders is installed.
    """


class IVendor(IDiscountSettingsEnabled):
    """A Vendor marker for containers.

    Used to create isolated areas within one shop for multiple vendors.
    """
